/** \file PhaseSpacePoint.hh
 *  \brief Contains the PhaseSpacePoint Class
 */

#pragma once

#include <vector>

#include "HEJ/Event.hh"
#include "HEJ/Particle.hh"
#include "HEJ/PDF.hh"
#include "HEJ/PDG_codes.hh"
#include "HEJ/RNG.hh"

#include "JetParameters.hh"
#include "ParticleProperties.hh"
#include "Status.hh"

namespace HEJFOG{
  class Process;

  using HEJ::Particle;
  //! A point in resummation phase space
  class PhaseSpacePoint{
  public:
    //! Default PhaseSpacePoint Constructor
    PhaseSpacePoint() = default;

    //! PhaseSpacePoint Constructor
    /**
     * @param proc                  The process to generate
     * @param jet_properties        Jet defintion & cuts
     * @param pdf                   The pdf set (used for sampling)
     * @param E_beam                Energie of the beam
     * @param subl_chance           Chance to turn a potentially unordered
     *                              emission into an actual one
     * @param subl_channels         Possible subleading channels.
     *                              see HEJFOG::Subleading
     * @param particle_properties   Properties of producted boson
     *
     * Initially, only FKL phase space points are generated. If the most
     * extremal emission in any direction is a quark or anti-quark and the
     * next emission is a gluon, subl_chance is the chance to turn this into
     * an unordered emission event by swapping the two flavours. At most one
     * unordered emission will be generated in this way.
     */
    PhaseSpacePoint(
      Process const & proc,
      JetParameters const & jet_properties,
      HEJ::PDF & pdf, double E_beam,
      double subl_chance,
      unsigned int subl_channels,
      ParticlesPropMap const & particles_properties,
      HEJ::RNG & ran
    );

    //! Get Weight Function
    /**
     * @returns        Weight of Event
     */
    double weight() const{
      return weight_;
    }

    Status status() const{
      return status_;
    }

    //! Get Incoming Function
    /**
     * @returns        Incoming Particles
     */
    std::array<Particle, 2> const & incoming() const{
      return incoming_;
    }

    //! Get Outgoing Function
    /**
     * @returns        Outgoing Particles
     */
    std::vector<Particle> const & outgoing() const{
      return outgoing_;
    }

    std::unordered_map<size_t, std::vector<Particle>> const & decays() const{
      return decays_;
    }

    private:
    /**
     * \internal
     * \brief Generate LO parton momentum
     *
     * @param count             Number of partons to generate
     * @param is_pure_jets      If true ensures momentum conservation in x and y
     * @param jet_param         Jet properties to fulfil
     * @param max_pt            max allowed pt for a parton (typically E_CMS)
     * @param ran               Random Number Generator
     *
     * @returns                 Momentum of partons
     *
     * Ensures that each parton is in its own jet.
     * Generation is independent of parton flavour. Output is sorted in rapidity.
     */
    std::vector<fastjet::PseudoJet> gen_LO_partons(
        int count, bool is_pure_jets,
        JetParameters const & jet_param,
        double max_pt,
        HEJ::RNG & ran
    );
    Particle gen_boson(
        HEJ::ParticleID bosonid, double mass, double width,
        HEJ::RNG & ran
    );
    template<class ParticleMomenta>
    fastjet::PseudoJet gen_last_momentum(
        ParticleMomenta const & other_momenta,
        double mass_square, double y
    ) const;

    bool jets_ok(
        std::vector<fastjet::PseudoJet> const & Born_jets,
        std::vector<fastjet::PseudoJet> const & partons
    ) const;
    /**
     * \internal
     * \brief Generate incoming partons according to the PDF
     *
     * @param uf                Scale used in the PDF
     */
    void reconstruct_incoming(
        std::array<HEJ::ParticleID, 2> const & ids,
        HEJ::PDF & pdf, double E_beam,
        double uf,
        HEJ::RNG & ran
    );
    HEJ::ParticleID generate_incoming_id(
        size_t beam_idx, double x, double uf, HEJ::PDF & pdf,
        HEJ::RNG & ran
    );

    bool momentum_conserved(double ep) const;

    HEJ::Particle const & most_backward_FKL(
        std::vector<HEJ::Particle> const & partons
    ) const;
    HEJ::Particle const & most_forward_FKL(
        std::vector<HEJ::Particle> const & partons
    ) const;
    HEJ::Particle & most_backward_FKL(std::vector<HEJ::Particle> & partons) const;
    HEJ::Particle & most_forward_FKL(std::vector<HEJ::Particle> & partons) const;
    bool extremal_FKL_ok(
        std::vector<fastjet::PseudoJet> const & partons
    ) const;
    double random_normal(double stddev, HEJ::RNG & ran);
    void maybe_turn_to_uno(double chance, HEJ::RNG & ran);
    std::vector<Particle> decay_boson(
        HEJ::Particle const & parent,
        std::vector<Decay> const & decays,
        HEJ::RNG & ran
    );
    Decay select_decay_channel(
        std::vector<Decay> const & decays,
        HEJ::RNG & ran
    );
    double gen_hard_pt(
        int np, double ptmin, double ptmax, double y,
        HEJ::RNG & ran
    );
    double gen_soft_pt(int np, double ptmax, HEJ::RNG & ran);
    double gen_parton_pt(
        int count, JetParameters const & jet_param, double ptmax, double y,
        HEJ::RNG & ran
    );


    double weight_;

    Status status_;

    std::array<Particle, 2> incoming_;
    std::vector<Particle> outgoing_;
    //! Particle decays in the format {outgoing index, decay products}
    std::unordered_map<size_t, std::vector<Particle>> decays_;
  };
  HEJ::UnclusteredEvent to_UnclusteredEvent(PhaseSpacePoint const & psp);
}
