#include "EventGenerator.hh"

#include "Process.hh"
#include "Beam.hh"
#include "JetParameters.hh"
#include "PhaseSpacePoint.hh"

#include "HEJ/Event.hh"
#include "HEJ/config.hh"

namespace HEJFOG{
  EventGenerator::EventGenerator(
      Process process,
      Beam beam,
      HEJ::ScaleGenerator scale_gen,
      JetParameters jets,
      int pdf_id,
      double subl_change,
      unsigned int subl_channels,
      ParticlesPropMap particles_properties,
      HEJ::HiggsCouplingSettings Higgs_coupling,
      HEJ::RNG & ran
  ):
    pdf_{pdf_id, beam.particles[0], beam.particles[1]},
    ME_{
      [this](double mu){ return pdf_.Halphas(mu); },
      HEJ::MatrixElementConfig{
        false,
        std::move(Higgs_coupling)
      }
    },
    scale_gen_{std::move(scale_gen)},
    process_{std::move(process)},
    jets_{std::move(jets)},
    beam_{std::move(beam)},
    subl_change_{subl_change},
    subl_channels_{subl_channels},
    particles_properties_{std::move(particles_properties)},
    ran_{ran}
  {
  }

  HEJ::Event EventGenerator::gen_event(){
    HEJFOG::PhaseSpacePoint psp{
      process_,
      jets_,
      pdf_, beam_.energy,
      subl_change_, subl_channels_,
      particles_properties_,
      ran_
    };
    status_ = psp.status();
    if(status_ != good) return {};

    HEJ::Event ev = scale_gen_(
        HEJ::Event{
          to_UnclusteredEvent(std::move(psp)),
          jets_.def, jets_.min_pt
        }
    );

    const double shat = HEJ::shat(ev);
    const double xa = (ev.incoming()[0].E()-ev.incoming()[0].pz())/(2.*beam_.energy);
    const double xb = (ev.incoming()[1].E()+ev.incoming()[1].pz())/(2.*beam_.energy);

    // evaluate matrix element on this point
    const auto ME_weights = ME_.tree(ev);
    ev.central().weight *= ME_weights.central/(shat*shat);
    ev.central().weight *= pdf_.pdfpt(0,xa,ev.central().muf, ev.incoming()[0].type);
    ev.central().weight *= pdf_.pdfpt(0,xb,ev.central().muf, ev.incoming()[1].type);
    for(size_t i = 0; i < ev.variations().size(); ++i){
      auto & var = ev.variations(i);
      var.weight *= ME_weights.variations[i]/(shat*shat);
      var.weight *= pdf_.pdfpt(0,xa,var.muf, ev.incoming()[0].type);
      var.weight *= pdf_.pdfpt(0,xb,var.muf, ev.incoming()[1].type);
    }
    return ev;
  }

}
