/** \file
 *  \brief Header file for the HepMCInterface
 *
 *  \authors   Jeppe Andersen, Tuomas Hapola, Marian Heil, Andreas Maier, Jennifer Smillie
 *  \date      2019
 *  \copyright GPLv2 or later
 */

#pragma once

#include <sys/types.h>
#include <vector>

namespace HepMC{
  class GenCrossSection;
  class GenEvent;
}

namespace HEJ{
  class Event;
  class EventParameters;
  //! This class converts the Events into HepMC::GenEvents
  /**
  *   \details The output is depended on the HepMC version HEJ is compiled with,
  *   both HepMC 2 and HepMC 3 are supported. If HEJ 2 is compiled
  *   without HepMC calling this interface will throw an error.
  *
  *   This interface will also keep track of the cross section of all the events that
  *   being fed into it.
  */

  class HepMCInterface{
  public:
    HepMCInterface();
    /**
     * \brief main function to convert an event into HepMC::GenEvent
     *
     * \param event          Event to convert
     * \param weight_index   optional selection of specific weight
     *                       (negative value gives central weight)
     */
    HepMC::GenEvent operator()(Event const & event, ssize_t weight_index = -1);
    /**
     * \brief initialise the event kinematics (everything but the weights)
     *
     * \param event          Event to convert
     * \param weight_index   optional selection of specific weight
     *                       (negative value gives central weight)
     */
    HepMC::GenEvent init_kinematics(Event const & event);
    /**
     * \brief Sets the central value from \p event to \p out_ev
     *
     * \param out_ev         HepMC::GenEvent to write to
     * \param event          Event to convert
     * \param weight_index   optional selection of specific weight
     *                       (negative value gives "central")
     */
    void set_central(HepMC::GenEvent & out_ev, Event const & event,
      ssize_t weight_index = -1);
    /**
     * \brief Add the event \p variations to \p out_ev
     */
    void add_variation(HepMC::GenEvent & out_ev,
      std::vector<EventParameters> const & variations);
  private:
    size_t event_count_;
    double tot_weight_;
    double tot_weight2_;
    HepMC::GenCrossSection cross_section() const;

  };
}
