//////////////////////////////////////////////////
//////////////////////////////////////////////////
// This source code is Copyright (2012) of      //
//  Jeppe R. Andersen and Jennifer M. Smillie   //
// and is distributed under the                 //
// Gnu Public License version 2                 //
// http://www.gnu.org/licenses/gpl-2.0.html     //
// You are allowed to distribute and alter the  //
// source under the conditions of the GPLv2     //
// as long as this copyright notice             //
// is unaltered and distributed with the source //
// Any use should comply with the               //
//             MCNET GUIDELINES                 //
//    for Event Generator Authors and Users     //
// as distributed with this source code         //
//////////////////////////////////////////////////
//////////////////////////////////////////////////

/** \file
 *  \brief Functions computing the square of current contractions.
 *
 *  This file contains all the necessary functions to compute the current
 *  contractions for all valid HEJ processes. PJETS, H+JETS and W+JETS along with
 *  some unordered counterparts.
 */

#pragma once

#include <complex>
#include <ostream>

#include <CLHEP/Vector/LorentzVector.h>

typedef std::complex<double> COM;
typedef COM current[4];
typedef CLHEP::HepLorentzVector HLV;

//! Square of qQ->qenuQ W+Jets Scattering Current
/**
 *  @param p1out                Momentum of final state quark
 *  @param pe                   Momentum of final state electron
 *  @param pnu                  Momentum of final state Neutrino
 *  @param p1in                 Momentum of initial state quark
 *  @param p2out                Momentum of final state quark
 *  @param p2in                 Momentum of intial state quark
 *  @returns                    Square of the current contractions for qQ->qenuQ Scattering
 *
 *  This returns the square of the current contractions in qQ->qenuQ scattering
 *  with an emission of a W Boson.
 */
double jMWqQ (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector pe,
              CLHEP::HepLorentzVector pnu,   CLHEP::HepLorentzVector p1in,
              CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in);


//! Square of qbarQ->qbarenuQ W+Jets Scattering Current
/**
 *  @param p1out                Momentum of final state anti-quark
 *  @param pe                   Momentum of final state electron
 *  @param pnu                  Momentum of final state Neutrino
 *  @param p1in                 Momentum of initial state anti-quark
 *  @param p2out                Momentum of final state quark
 *  @param p2in                 Momentum of intial state quark
 *  @returns                    Square of the current contractions for qbarQ->qbarenuQ Scattering
 *
 *  This returns the square of the current contractions in qbarQ->qbarenuQ scattering
 *  with an emission of a W Boson.
 */
double jMWqbarQ (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector pe,
                 CLHEP::HepLorentzVector pnu, CLHEP::HepLorentzVector p1in,
                 CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in);


//! Square of qQbar->qenuQbar W+Jets Scattering Current
/**
 *  @param p1out                Momentum of final state quark
 *  @param pe                   Momentum of final state electron
 *  @param pnu                  Momentum of final state Neutrino
 *  @param p1in                 Momentum of initial state quark
 *  @param p2out                Momentum of final state anti-quark
 *  @param p2in                 Momentum of intial state anti-quark
 *  @returns                    Square of the current contractions for qQbar->qenuQbar Scattering
 *
 *  This returns the square of the current contractions in qQbar->qenuQbar scattering
 *  with an emission of a W Boson.
 */
double jMWqQbar (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector pe,
                 CLHEP::HepLorentzVector pnu, CLHEP::HepLorentzVector p1in,
                 CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in);


//! Square of qbarQbar->qbarenuQbar W+Jets Scattering Current
/**
 *  @param p1out                Momentum of final state anti-quark
 *  @param pe                   Momentum of final state electron
 *  @param pnu                  Momentum of final state Neutrino
 *  @param p1in                 Momentum of initial state anti-quark
 *  @param p2out                Momentum of final state anti-quark
 *  @param p2in                 Momentum of intial state anti-quark
 *  @returns                    Square of the current contractions for qbarQbar->qbarenuQbar Scattering
 *
 *  This returns the square of the current contractions in qbarQbar->qbarenuQbar scattering
 *  with an emission of a W Boson.
 */
double jMWqbarQbar (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector pe,
                    CLHEP::HepLorentzVector pnu, CLHEP::HepLorentzVector p1in,
                    CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in);

//! Square of qg->qenug W+Jets Scattering Current
/**
 *  @param p1out                Momentum of final state quark
 *  @param pe                   Momentum of final state electron
 *  @param pnu                  Momentum of final state Neutrino
 *  @param p1in                 Momentum of initial state quark
 *  @param p2out                Momentum of final state gluon
 *  @param p2in                 Momentum of intial state gluon
 *  @returns                    Square of the current contractions for qg->qenug Scattering
 *
 *  This returns the square of the current contractions in qg->qenug scattering
 *  with an emission of a W Boson.
 */
double jMWqg (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector pe,
              CLHEP::HepLorentzVector pnu, CLHEP::HepLorentzVector p1in,
              CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in);

//! Square of qbarg->qbarenug W+Jets Scattering Current
/**
 *  @param p1out                Momentum of final state anti-quark
 *  @param pe                   Momentum of final state electron
 *  @param pnu                  Momentum of final state Neutrino
 *  @param p1in                 Momentum of initial state anti-quark
 *  @param p2out                Momentum of final state gluon
 *  @param p2in                 Momentum of intial state gluon
 *  @returns                    Square of the current contractions for qbarg->qbarenug Scattering
 *
 *  This returns the square of the current contractions in qbarg->qbarenug scattering
 *  with an emission of a W Boson.
 */
double jMWqbarg (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector pe,
                 CLHEP::HepLorentzVector pnu, CLHEP::HepLorentzVector p1in,
                 CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in);


//! Square of qQ->qQ Pure Jets Scattering Current
/**
 *  @param p1out                Momentum of final state quark
 *  @param p1in                 Momentum of initial state quark
 *  @param p2out                Momentum of final state quark
 *  @param p2in                 Momentum of intial state quark
 *  @returns                    Square of the current contractions for qQ->qQ Scattering
 *
 *  This returns the square of the current contractions in qQ->qQ Pure Jet Scattering.
 */
double jM2qQ (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in,
              CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in);


//! Square of qQbar->qQbar Pure Jets Scattering Current
/**
 *  @param p1out                Momentum of final state quark
 *  @param p1in                 Momentum of initial state quark
 *  @param p2out                Momentum of final state anti-quark
 *  @param p2in                 Momentum of intial state anti-quark
 *  @returns                    Square of the current contractions for qQbar->qQbar Scattering
 *
 *  This returns the square of the current contractions in qQbar->qQbar Pure Jet Scattering.
 *  Note this can be used for qbarQ->qbarQ Scattering by inputting arguments appropriately.
 */
double jM2qQbar (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in,
                 CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in);


//! Square of qbarQbar->qbarQbar Pure Jets Scattering Current
/**
 *  @param p1out                Momentum of final state anti-quark
 *  @param p1in                 Momentum of initial state anti-quark
 *  @param p2out                Momentum of final state anti-quark
 *  @param p2in                 Momentum of intial state anti-quark
 *  @returns                    Square of the current contractions for qbarQbar->qbarQbar Scattering
 *
 *  This returns the square of the current contractions in qbarQbar->qbarQbar Pure Jet Scattering.
 */
double jM2qbarQbar (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in,
                    CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in);


//! Square of qg->qg Pure Jets Scattering Current
/**
 *  @param p1out                Momentum of final state quark
 *  @param p1in                 Momentum of initial state quark
 *  @param p2out                Momentum of final state gluon
 *  @param p2in                 Momentum of intial state gluon
 *  @returns                    Square of the current contractions for qg->qg Scattering
 *
 *  This returns the square of the current contractions in qg->qg Pure Jet Scattering.
 *  Note this can be used for gq->gq Scattering by inputting arguments appropriately.
 */
double jM2qg (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in,
              CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in);


//! Square of qbarg->qbarg Pure Jets Scattering Current
/**
 *  @param p1out                Momentum of final state anti-quark
 *  @param p1in                 Momentum of initial state anti-quark
 *  @param p2out                Momentum of final state gluon
 *  @param p2in                 Momentum of intial state gluon
 *  @returns                    Square of the current contractions for qbarg->qbarg Scattering
 *
 *  This returns the square of the current contractions in qbarg->qbarg Pure Jet Scattering.
 *  Note this can be used for gqbar->gqbar Scattering by inputting arguments appropriately.
 */
double jM2qbarg (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in,
                 CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in);


//! Square of gg->gg Pure Jets Scattering Current
/**
 *  @param p1out                Momentum of final state gluon
 *  @param p1in                 Momentum of initial state gluon
 *  @param p2out                Momentum of final state gluon
 *  @param p2in                 Momentum of intial state gluon
 *  @returns                    Square of the current contractions for gg->gg Scattering
 *
 *  This returns the square of the current contractions in gg->gg Pure Jet Scattering.
 */
double jM2gg (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in,
              CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in);


//! Square of gg->gg Higgs+Jets Scattering Current
/**
 *  @param p1out                Momentum of final state gluon
 *  @param p1in                 Momentum of initial state gluon
 *  @param p2out                Momentum of final state gluon
 *  @param p2in                 Momentum of intial state gluon
 *  @param q1                   Momentum of t-channel propagator before Higgs
 *  @param qH2                  Momentum of t-channel propagator after Higgs
 *  @param mt                   Top quark mass
 *  @param include_bottom       Specifies whether bottom corrections are included
 *  @param mb                   Bottom quark mass
 *  @returns                    Square of the current contractions for gg->gg Scattering
 *
 *  This returns the square of the current contractions in gg->gg Higgs+Jet Scattering.
 *
 *  g~p1 g~p2
 *  should be called with q1 meant to be contracted with p2 in first part of vertex
 *  (i.e. if g is backward, q1 is forward)
 */
double MH2gg (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in,
              CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in,
              CLHEP::HepLorentzVector q1, CLHEP::HepLorentzVector qH2,
              double mt,
              bool include_bottom, double mb);

//! Square of gq->gq Higgs+Jets Scattering Current with Higgs before Gluon
/**
 *  @param p1out                Momentum of final state gluon
 *  @param p1in                 Momentum of initial state gluon
 *  @param p2out                Momentum of final state gluon
 *  @param p2in                 Momentum of intial state gluon
 *  @param pH                   Momentum of Higgs
 *  @param mt                   Top quark mass
 *  @param include_bottom       Specifies whether bottom corrections are included
 *  @param mb                   Bottom quark mass
 *  @returns                    Square of the current contraction
 *
 */
double MH2gq_outsideH(CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in,
                      CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in,
                      CLHEP::HepLorentzVector pH,
                      double mt,
                      bool include_bottom, double mb);


//! Square of qg->qg Higgs+Jets Scattering Current
/**
 *  @param p1out                Momentum of final state quark
 *  @param p1in                 Momentum of initial state quark
 *  @param p2out                Momentum of final state gluon
 *  @param p2in                 Momentum of intial state gluon
 *  @param q1                   Momentum of t-channel propagator before Higgs
 *  @param qH2                  Momentum of t-channel propagator after Higgs
 *  @param mt                   Top quark mass
 *  @param include_bottom       Specifies whether bottom corrections are included
 *  @param mb                   Bottom quark mass
 *  @returns                    Square of the current contractions for qg->qg Scattering
 *
 *  This returns the square of the current contractions in qg->qg Higgs+Jet Scattering.
 *
 *  q~p1 g~p2 (i.e. ALWAYS p1 for quark, p2 for gluon)
 *  should be called with q1 meant to be contracted with p2 in first part of vertex
 *  (i.e. if g is backward, q1 is forward)
 */
double MH2qg (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in,
              CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in,
              CLHEP::HepLorentzVector q1, CLHEP::HepLorentzVector qH2,
              double mt,
              bool include_bottom, double mb);


//! Square of qbarg->qbarg Higgs+Jets Scattering Current
/**
 *  @param p1out                Momentum of final state anti-quark
 *  @param p1in                 Momentum of initial state anti-quark
 *  @param p2out                Momentum of final state gluon
 *  @param p2in                 Momentum of intial state gluon
 *  @param q1                   Momentum of t-channel propagator before Higgs
 *  @param qH2                  Momentum of t-channel propagator after Higgs
 *  @param mt                   Top quark mass
 *  @param include_bottom       Specifies whether bottom corrections are included
 *  @param mb                   Bottom quark mass
 *  @returns                    Square of the current contractions for qbarg->qbarg Scattering
 *
 *  This returns the square of the current contractions in qbarg->qbarg Higgs+Jet Scattering.
 *
 *  qbar~p1 g~p2 (i.e. ALWAYS p1 for anti-quark, p2 for gluon)
 *  should be called with q1 meant to be contracted with p2 in first part of vertex
 *  (i.e. if g is backward, q1 is forward)
 */
double MH2qbarg (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in,
                 CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in,
                 CLHEP::HepLorentzVector q1, CLHEP::HepLorentzVector qH2,
                 double mt,
                 bool include_bottom, double mb);


//! Square of qQ->qQ Higgs+Jets Scattering Current
/**
 *  @param p1out                Momentum of final state quark
 *  @param p1in                 Momentum of initial state quark
 *  @param p2out                Momentum of final state quark
 *  @param p2in                 Momentum of intial state quark
 *  @param q1                   Momentum of t-channel propagator before Higgs
 *  @param qH2                  Momentum of t-channel propagator after Higgs
 *  @param mt                   Top quark mass
 *  @param include_bottom       Specifies whether bottom corrections are included
 *  @param mb                   Bottom quark mass
 *  @returns                    Square of the current contractions for qQ->qQ Scattering
 *
 *  This returns the square of the current contractions in qQ->qQ Higgs+Jet Scattering.
 *
 *  q~p1 Q~p2 (i.e. ALWAYS p1 for quark, p2 for quark)
 *  should be called with q1 meant to be contracted with p2 in first part of vertex
 *  (i.e. if Q is backward, q1 is forward)
 */
double MH2qQ (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in,
              CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in,
              CLHEP::HepLorentzVector q1, CLHEP::HepLorentzVector qH2,
              double mt,
              bool include_bottom, double mb);

//! Square of qQbar->qQbar Higgs+Jets Scattering Current
/**
 *  @param p1out                Momentum of final state quark
 *  @param p1in                 Momentum of initial state quark
 *  @param p2out                Momentum of final state anti-quark
 *  @param p2in                 Momentum of intial state anti-quark
 *  @param q1                   Momentum of t-channel propagator before Higgs
 *  @param qH2                   Momentum of t-channel propagator after Higgs
 *  @param mt                   Top quark mass
 *  @param include_bottom       Specifies whether bottom corrections are included
 *  @param mb                   Bottom quark mass
 *  @returns                    Square of the current contractions for qQ->qQ Scattering
 *
 *  This returns the square of the current contractions in qQbar->qQbar Higgs+Jet Scattering.
 *
 *  q~p1 Qbar~p2 (i.e. ALWAYS p1 for quark, p2 for anti-quark)
 *  should be called with q1 meant to be contracted with p2 in first part of vertex
 *  (i.e. if Qbar is backward, q1 is forward)
 */
double MH2qQbar (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in,
                 CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in,
                 CLHEP::HepLorentzVector q1, CLHEP::HepLorentzVector qH2,
                 double mt,
                 bool include_bottom, double mb);

//! Square of qbarQ->qbarQ Higgs+Jets Scattering Current
/**
 *  @param p1out                Momentum of final state anti-quark
 *  @param p1in                 Momentum of initial state anti-quark
 *  @param p2out                Momentum of final state quark
 *  @param p2in                 Momentum of intial state quark
 *  @param q1                   Momentum of t-channel propagator before Higgs
 *  @param qH2                  Momentum of t-channel propagator after Higgs
 *  @param mt                   Top quark mass
 *  @param include_bottom       Specifies whether bottom corrections are included
 *  @param mb                   Bottom quark mass
 *  @returns                    Square of the current contractions for qbarQ->qbarQ Scattering
 *
 *  This returns the square of the current contractions in qbarQ->qbarQ Higgs+Jet Scattering.
 *
 *  qbar~p1 Q~p2 (i.e. ALWAYS p1 for anti-quark, p2 for quark)
 *  should be called with q1 meant to be contracted with p2 in first part of vertex
 *  (i.e. if Q is backward, q1 is forward)
 */
double MH2qbarQ (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in,
                 CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in,
                 CLHEP::HepLorentzVector q1, CLHEP::HepLorentzVector qH2,
                 double mt,
                 bool include_bottom, double mb);


//! Square of qbarQbar->qbarQbar Higgs+Jets Scattering Current
/**
 *  @param p1out                Momentum of final state anti-quark
 *  @param p1in                 Momentum of initial state anti-quark
 *  @param p2out                Momentum of final state anti-quark
 *  @param p2in                 Momentum of intial state anti-quark
 *  @param q1                   Momentum of t-channel propagator before Higgs
 *  @param qH2                  Momentum of t-channel propagator after Higgs
 *  @param mt                   Top quark mass
 *  @param include_bottom       Specifies whether bottom corrections are included
 *  @param mb                   Bottom quark mass
 *  @returns                    Square of the current contractions for qbarQbar->qbarQbar Scattering
 *
 *  This returns the square of the current contractions in qbarQbar->qbarQbar Higgs+Jet Scattering.
 *
 *  qbar~p1 Qbar~p2 (i.e. ALWAYS p1 for anti-quark, p2 for anti-quark)
 *  should be called with q1 meant to be contracted with p2 in first part of vertex
 *  (i.e. if Qbar is backward, q1 is forward)
 */
double MH2qbarQbar (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in,
                    CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in,
                    CLHEP::HepLorentzVector q1, CLHEP::HepLorentzVector qH2,
                    double mt,
                    bool include_bottom, double mb);


// Unordered f

//! Square of qQ->gqQ Higgs+Jets Unordered f Scattering Current
/**
 *  @param pg                   Momentum of unordered gluon
 *  @param p1out                Momentum of final state quark
 *  @param p1in                 Momentum of initial state quark
 *  @param p2out                Momentum of final state quark
 *  @param p2in                 Momentum of intial state quark
 *  @param qH1                  Momentum of t-channel propagator before Higgs
 *  @param qH2                  Momentum of t-channel propagator after Higgs
 *  @param mt                   Top quark mass
 *  @param include_bottom       Specifies whether bottom corrections are included
 *  @param mb                   Bottom quark mass
 *  @returns                    Square of the current contractions for qQ->gqQ Scattering
 *
 *  This returns the square of the current contractions in qQ->gqQ Higgs+Jet Scattering.
 *
 *  This construction is taking rapidity order: pg > p1out >> p2out
 */
double jM2unogqHQ (CLHEP::HepLorentzVector pg, CLHEP::HepLorentzVector p1out,
                   CLHEP::HepLorentzVector p1in, CLHEP::HepLorentzVector p2out,
                   CLHEP::HepLorentzVector p2in, CLHEP::HepLorentzVector qH1,
                   CLHEP::HepLorentzVector qH2,
                   double mt,
                   bool include_bottom, double mb);


//! Square of qQbar->gqQbar Higgs+Jets Unordered f Scattering Current
/**
 *  @param pg                   Momentum of unordered gluon
 *  @param p1out                Momentum of final state quark
 *  @param p1in                 Momentum of initial state quark
 *  @param p2out                Momentum of final state anti-quark
 *  @param p2in                 Momentum of intial state anti-quark
 *  @param qH1                  Momentum of t-channel propagator before Higgs
 *  @param qH2                  Momentum of t-channel propagator after Higgs
 *  @param mt                   Top quark mass
 *  @param include_bottom       Specifies whether bottom corrections are included
 *  @param mb                   Bottom quark mass
 *  @returns                    Square of the current contractions for qQbar->gqQbar Scattering
 *
 *  This returns the square of the current contractions in qQbar->gqQbar Higgs+Jet Scattering.
 *
 *  This construction is taking rapidity order: pg > p1out >> p2out
 */
double jM2unogqHQbar (CLHEP::HepLorentzVector pg, CLHEP::HepLorentzVector p1out,
                      CLHEP::HepLorentzVector p1in, CLHEP::HepLorentzVector p2out,
                      CLHEP::HepLorentzVector p2in, CLHEP::HepLorentzVector qH1,
                      CLHEP::HepLorentzVector qH2,
                      double mt,
                      bool include_bottom, double mb);

//! Square of qbarQ->gqbarQ Higgs+Jets Unordered f Scattering Current
/**
 *  @param pg                   Momentum of unordered gluon
 *  @param p1out                Momentum of final state anti-quark
 *  @param p1in                 Momentum of initial state anti-quark
 *  @param p2out                Momentum of final state quark
 *  @param p2in                 Momentum of intial state quark
 *  @param qH1                  Momentum of t-channel propagator before Higgs
 *  @param qH2                  Momentum of t-channel propagator after Higgs
 *  @param mt                   Top quark mass
 *  @param include_bottom       Specifies whether bottom corrections are included
 *  @param mb                   Bottom quark mass
 *  @returns                    Square of the current contractions for qbarQ->gqbarQ Scattering
 *
 *  This returns the square of the current contractions in qbarQ->gqbarQ Higgs+Jet Scattering.
 *
 *  This construction is taking rapidity order: pg > p1out >> p2out
 */
double jM2unogqbarHQ (CLHEP::HepLorentzVector pg, CLHEP::HepLorentzVector p1out,
                      CLHEP::HepLorentzVector p1in, CLHEP::HepLorentzVector p2out,
                      CLHEP::HepLorentzVector p2in, CLHEP::HepLorentzVector qH1,
                      CLHEP::HepLorentzVector qH2,
                      double mt,
                      bool include_bottom, double mb);

//! Square of qbarQbar->gqbarQbar Higgs+Jets Unordered f Scattering Current
/**
 *  @param pg                   Momentum of unordered gluon
 *  @param p1out                Momentum of final state anti-quark
 *  @param p1in                 Momentum of initial state anti-quark
 *  @param p2out                Momentum of final state anti-quark
 *  @param p2in                 Momentum of intial state anti-quark
 *  @param qH1                  Momentum of t-channel propagator before Higgs
 *  @param qH2                  Momentum of t-channel propagator after Higgs
 *  @param mt                   Top quark mass
 *  @param include_bottom       Specifies whether bottom corrections are included
 *  @param mb                   Bottom quark mass
 *  @returns                    Square of the current contractions for qbarQbar->gqbarQbar Scattering
 *
 *  This returns the square of the current contractions in qbarQbar->gqbarQbar Higgs+Jet Scattering.
 *
 *  This construction is taking rapidity order: pg > p1out >> p2out
 */
double jM2unogqbarHQbar (CLHEP::HepLorentzVector pg, CLHEP::HepLorentzVector p1out,
                         CLHEP::HepLorentzVector p1in, CLHEP::HepLorentzVector p2out,
                         CLHEP::HepLorentzVector p2in, CLHEP::HepLorentzVector qH1,
                         CLHEP::HepLorentzVector qH2,
                         double mt,
                         bool include_bottom, double mb);


//! Square of qg->gqg Higgs+Jets Unordered f Scattering Current
/**
 *  @param pg                   Momentum of unordered gluon
 *  @param p1out                Momentum of final state quark
 *  @param p1in                 Momentum of initial state quark
 *  @param p2out                Momentum of final state gluon
 *  @param p2in                 Momentum of intial state gluon
 *  @param qH1                  Momentum of t-channel propagator before Higgs
 *  @param qH2                  Momentum of t-channel propagator after Higgs
 *  @param mt                   Top quark mass
 *  @param include_bottom       Specifies whether bottom corrections are included
 *  @param mb                   Bottom quark mass
 *  @returns                    Square of the current contractions for qg->gqg Scattering
 *
 *  This returns the square of the current contractions in qg->gqg Higgs+Jet Scattering.
 *
 *  This construction is taking rapidity order: pg > p1out >> p2out
 */
double jM2unogqHg (CLHEP::HepLorentzVector pg, CLHEP::HepLorentzVector p1out,
                   CLHEP::HepLorentzVector p1in, CLHEP::HepLorentzVector p2out,
                   CLHEP::HepLorentzVector p2in, CLHEP::HepLorentzVector qH1,
                   CLHEP::HepLorentzVector qH2,
                   double mt,
                   bool include_bottom, double mb);


//! Square of qbarg->gqbarg Higgs+Jets Unordered f Scattering Current
/**
 *  @param pg                   Momentum of unordered gluon
 *  @param p1out                Momentum of final state anti-quark
 *  @param p1in                 Momentum of initial state anti-quark
 *  @param p2out                Momentum of final state gluon
 *  @param p2in                 Momentum of intial state gluon
 *  @param qH1                  Momentum of t-channel propagator before Higgs
 *  @param qH2                  Momentum of t-channel propagator after Higgs
 *  @param mt                   Top quark mass
 *  @param include_bottom       Specifies whether bottom corrections are included
 *  @param mb                   Bottom quark mass
 *  @returns                    Square of the current contractions for qbarg->gbarg Scattering
 *
 *  This returns the square of the current contractions in qbarg->gqbarg Higgs+Jet Scattering.
 *
 *  This construction is taking rapidity order: pg > p1out >> p2out
 */
double jM2unogqbarHg (CLHEP::HepLorentzVector pg, CLHEP::HepLorentzVector p1out,
                      CLHEP::HepLorentzVector p1in, CLHEP::HepLorentzVector p2out,
                      CLHEP::HepLorentzVector p2in, CLHEP::HepLorentzVector qH1,
                      CLHEP::HepLorentzVector qH2,
                      double mt,
                      bool include_bottom, double mb);


//Unordered b

//! Square of qbarQ->qbarQg Higgs+Jets Unordered b Scattering Current
/**
 *  @param p1out                Momentum of final state anti-quark
 *  @param p1in                 Momentum of initial state anti-quark
 *  @param pg                   Momentum of unordered b gluon
 *  @param p2out                Momentum of final state quark
 *  @param p2in                 Momentum of intial state quark
 *  @param qH1                  Momentum of t-channel propagator before Higgs
 *  @param qH2                  Momentum of t-channel propagator after Higgs
 *  @param mt                   Top quark mass
 *  @param include_bottom       Specifies whether bottom corrections are included
 *  @param mb                   Bottom quark mass
 *  @returns                    Square of the current contractions for qbarQ->qbarQg Scattering
 *
 *  This returns the square of the current contractions in qbarQ->qbarQg Higgs+Jet Scattering.
 *
 *  This construction is taking rapidity order: p1out >> p2out > pg
 */
double jM2unobqbarHQg (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in,
                       CLHEP::HepLorentzVector pg, CLHEP::HepLorentzVector p2out,
                       CLHEP::HepLorentzVector p2in, CLHEP::HepLorentzVector qH1,
                       CLHEP::HepLorentzVector qH2,
                       double mt,
                       bool include_bottom, double mb);


//! Square of qQ->qQg Higgs+Jets Unordered b Scattering Current
/**
 *  @param p1out                Momentum of final state quark
 *  @param p1in                 Momentum of initial state quark
 *  @param pg                   Momentum of unordered b gluon
 *  @param p2out                Momentum of final state quark
 *  @param p2in                 Momentum of intial state quark
 *  @param qH1                  Momentum of t-channel propagator before Higgs
 *  @param qH2                  Momentum of t-channel propagator after Higgs
 *  @param mt                   Top quark mass
 *  @param include_bottom       Specifies whether bottom corrections are included
 *  @param mb                   Bottom quark mass
 *  @returns                    Square of the current contractions for qQ->qQg Scattering
 *
 *  This returns the square of the current contractions in qQ->qQg Higgs+Jet Scattering.
 *
 *  This construction is taking rapidity order: p1out >> p2out > pg
 */
double jM2unobqHQg (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in,
                    CLHEP::HepLorentzVector pg, CLHEP::HepLorentzVector p2out,
                    CLHEP::HepLorentzVector p2in, CLHEP::HepLorentzVector qH1,
                    CLHEP::HepLorentzVector qH2,
                    double mt,
                    bool include_bottom, double mb);


//! Square of qQbar->qQbarg Higgs+Jets Unordered b Scattering Current
/**
 *  @param p1out                Momentum of final state quark
 *  @param p1in                 Momentum of initial state quark
 *  @param pg                   Momentum of unordered b gluon
 *  @param p2out                Momentum of final state anti-quark
 *  @param p2in                 Momentum of intial state anti-quark
 *  @param qH1                  Momentum of t-channel propagator before Higgs
 *  @param qH2                  Momentum of t-channel propagator after Higgs
 *  @param mt                   Top quark mass
 *  @param include_bottom       Specifies whether bottom corrections are included
 *  @param mb                   Bottom quark mass
 *  @returns                    Square of the current contractions for qQbar->qQbarg Scattering
 *
 *  This returns the square of the current contractions in qQbar->qQbarg Higgs+Jet Scattering.
 *
 *  This construction is taking rapidity order: p1out >> p2out > pg
 */
double jM2unobqHQbarg (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in,
                       CLHEP::HepLorentzVector pg, CLHEP::HepLorentzVector p2out,
                       CLHEP::HepLorentzVector p2in, CLHEP::HepLorentzVector qH1,
                       CLHEP::HepLorentzVector qH2,
                       double mt,
                       bool include_bottom, double mb);

//! Square of qbarQbar->qbarQbarg Higgs+Jets Unordered b Scattering Current
/**
 *  @param p1out                Momentum of final state anti-quark
 *  @param p1in                 Momentum of initial state anti-quark
 *  @param pg                   Momentum of unordered b gluon
 *  @param p2out                Momentum of final state anti-quark
 *  @param p2in                 Momentum of intial state anti-quark
 *  @param qH1                  Momentum of t-channel propagator before Higgs
 *  @param qH2                  Momentum of t-channel propagator after Higgs
 *  @param mt                   Top quark mass
 *  @param include_bottom       Specifies whether bottom corrections are included
 *  @param mb                   Bottom quark mass
 *  @returns                    Square of the current contractions for qbarQbar->qbarQbarg Scattering
 *
 *  This returns the square of the current contractions in qbarQbar->qbarQbarg Higgs+Jet Scattering.
 *
 *  This construction is taking rapidity order: p1out >> p2out > pg
 */
double jM2unobqbarHQbarg (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in,
                          CLHEP::HepLorentzVector pg, CLHEP::HepLorentzVector p2out,
                          CLHEP::HepLorentzVector p2in, CLHEP::HepLorentzVector qH1,
                          CLHEP::HepLorentzVector qH2,
                          double mt,
                          bool include_bottom, double mb);


//! Square of gQbar->gQbarg Higgs+Jets Unordered b Scattering Current
/**
 *  @param p1out                Momentum of final state gluon
 *  @param p1in                 Momentum of initial state gluon
 *  @param pg                   Momentum of unordered b gluon
 *  @param p2out                Momentum of final state anti-quark
 *  @param p2in                 Momentum of intial state anti-quark
 *  @param qH1                  Momentum of t-channel propagator before Higgs
 *  @param qH2                  Momentum of t-channel propagator after Higgs
 *  @param mt                   Top quark mass
 *  @param include_bottom       Specifies whether bottom corrections are included
 *  @param mb                   Bottom quark mass
 *  @returns                    Square of the current contractions for gQbar->gQbarg Scattering
 *
 *  This returns the square of the current contractions in gQbar->gQbarg Higgs+Jet Scattering.
 *
 *  This construction is taking rapidity order: p1out >> p2out > pg
 */
double jM2unobgHQbarg (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in,
                       CLHEP::HepLorentzVector pg, CLHEP::HepLorentzVector p2out,
                       CLHEP::HepLorentzVector p2in, CLHEP::HepLorentzVector qH1,
                       CLHEP::HepLorentzVector qH2,
                       double mt,
                       bool include_bottom, double mb);

//! Square of gQ->gQg Higgs+Jets Unordered b Scattering Current
/**
 *  @param p1out                Momentum of final state gluon
 *  @param p1in                 Momentum of initial state gluon
 *  @param pg                   Momentum of unordered b gluon
 *  @param p2out                Momentum of final state quark
 *  @param p2in                 Momentum of intial state quark
 *  @param qH1                  Momentum of t-channel propagator before Higgs
 *  @param qH2                  Momentum of t-channel propagator after Higgs
 *  @param mt                   Top quark mass
 *  @param include_bottom       Specifies whether bottom corrections are included
 *  @param mb                   Bottom quark mass
 *  @returns                    Square of the current contractions for gQ->gQg Scattering
 *
 *  This returns the square of the current contractions in gQ->gQg Higgs+Jet Scattering.
 *
 *  This construction is taking rapidity order: p1out >> p2out > pg
 */
double jM2unobgHQg (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in,
                    CLHEP::HepLorentzVector pg, CLHEP::HepLorentzVector p2out,
                    CLHEP::HepLorentzVector p2in, CLHEP::HepLorentzVector qH1,
                    CLHEP::HepLorentzVector qH2,
                    double mt,
                    bool include_bottom, double mb);

// impact factors for Higgs + jet


//! Implements Eq. (4.22) in hep-ph/0301013 with modifications to incoming plus momenta
/**
 * @param p2               Momentum of Particle 2
 * @param p1               Momentum of Particle 1
 * @param pH               Momentum of Higgs
 * @returns                Value of Eq. (4.22) in Hep-ph/0301013 with modifications
 *
 *  This gives the impact factor. First it determines first whether this is the case
 *  p1p\sim php>>p3p or the opposite
 */
double C2gHgm(CLHEP::HepLorentzVector p2, CLHEP::HepLorentzVector p1,
              CLHEP::HepLorentzVector pH);


//! Implements Eq. (4.23) in hep-ph/0301013 with modifications to incoming plus momenta
/**
 * @param p2               Momentum of Particle 2
 * @param p1               Momentum of Particle 1
 * @param pH               Momentum of Higgs
 * @returns                Value of Eq. (4.23) in Hep-ph/0301013
 *
 *  This gives the impact factor. First it determines first whether this is the case
 *  p1p\sim php>>p3p or the opposite
 */
double C2gHgp(CLHEP::HepLorentzVector p2, CLHEP::HepLorentzVector p1,
              CLHEP::HepLorentzVector pH);


//! Implements Eq. (4.22) in hep-ph/0301013
/**
 * @param p2               Momentum of Particle 2
 * @param p1               Momentum of Particle 1
 * @param pH               Momentum of Higgs
 * @returns                Value of Eq. (4.22) in Hep-ph/0301013
 *
 *  This gives the impact factor. First it determines first whether this is the case
 *  p1p\sim php>>p3p or the opposite
 */
double C2qHqm(CLHEP::HepLorentzVector p2, CLHEP::HepLorentzVector p1,
              CLHEP::HepLorentzVector pH);

/** \class CCurrent currents.hh "include/HEJ/currents.hh"
 *  \brief This is the a new class structure for currents.
 */
class CCurrent
{
public:
    CCurrent(COM sc0, COM sc1, COM sc2, COM sc3)
    :c0(sc0),c1(sc1),c2(sc2),c3(sc3)
    {};
    CCurrent(const CLHEP::HepLorentzVector p)
    {
        c0=p.e();
        c1=p.px();
        c2=p.py();
        c3=p.pz();
    };
    CCurrent()
    {};
    CCurrent operator+(const CCurrent& other);
    CCurrent operator-(const CCurrent& other);
    CCurrent operator*(const double x);
    CCurrent operator*(const COM x);
    CCurrent operator/(const double x);
    CCurrent operator/(const COM x);

    friend std::ostream& operator<<(std::ostream& os, const CCurrent& cur);
    COM dot(CLHEP::HepLorentzVector p1);
    COM dot(CCurrent p1);
    COM c0,c1,c2,c3;
private:
};

/* std::ostream& operator <<(std::ostream& os, const CCurrent& cur); */
CCurrent operator * ( double x, CCurrent& m);
CCurrent operator * ( COM x, CCurrent& m);
CCurrent operator / ( double x, CCurrent& m);
CCurrent operator / ( COM x, CCurrent& m);

//! Current <outgoing state | mu | incoming state>
/**
 * These functions are a mess. There are many more defined in the source file than declared in the
 * header - and the arguments are mislabelled in some cases. Need to investigate.
 */
//! @TODO remove
[[deprecated("Use joi instead")]]
CCurrent j (CLHEP::HepLorentzVector pout, bool helout, CLHEP::HepLorentzVector pin, bool helin);

//! Current <outgoing state | mu | incoming state>
/**
 * These functions are a mess. There are many more defined in the source file than declared in the
 * header - and the arguments are mislabelled in some cases. Need to investigate.
 */
CCurrent joi (CLHEP::HepLorentzVector pout, bool helout, CLHEP::HepLorentzVector pin, bool helin);

//! Current <incoming state | mu | outgoing state>
/**
 * These functions are a mess. There are many more defined in the source file than declared in the
 * header - and the arguments are mislabelled in some cases. Need to investigate.
 */
CCurrent jio (CLHEP::HepLorentzVector pout, bool helout, CLHEP::HepLorentzVector pin, bool helin);

//! Current <outgoing state | mu | outgoing state>
/**
 * These functions are a mess. There are many more defined in the source file than declared in the
 * header - and the arguments are mislabelled in some cases. Need to investigate.
 */
CCurrent joo (CLHEP::HepLorentzVector pout, bool helout, CLHEP::HepLorentzVector pin, bool helin);

/* // Coupling values */
/* const double stw2   = 0.2222; */
/* const double ctw    = sqrt(1.0 - stw2); */
/* const double gs     = 1.217716; */
/* const double gw     = 0.653232911; */
/* const double Zem    = (-1.0 / 2.0 + stw2) / ctw; */
/* const double Zep    = stw2 / ctw; */
/* const double Zum    = ( 1.0 / 2.0 - 2.0 * stw2 / 3.0) / ctw; */
/* const double Zup    = - 2.0 * stw2 / 3.0 / ctw; */
/* const double Zdm    = (-1.0 / 2.0 + 1.0 / 3.0 * stw2) / ctw; */
/* const double Zdp    = stw2 / 3.0 / ctw; */
/* const double RWeak  = -pow(gw, 2.0); */
/* const double Strong =  pow(gs, 4.0); */
/* const double ee     =  pow(gw, 2.0) * stw2; */


/* std::vector <double>  jMZqQ (HLV, HLV, HLV, HLV, HLV, HLV, std::vector <double>, std::vector < std::vector <double> >, int, int, bool, bool); */
/* std::vector <double>  jMZqg (HLV, HLV, HLV, HLV, HLV, HLV, std::vector <double>, std::vector < std::vector <double> >, int, int, bool, bool); */
/* void jZ    (HLV, HLV, HLV, HLV, bool, bool, current); */
/* void jZbar (HLV, HLV, HLV, HLV, bool, bool, current); */
/* COM PZ(double); */
/* double Zq (int, bool); */
/* double Gq (int); */
