/**
 * \file
 * \brief Contains various utilities
 *
 *  \authors   Jeppe Andersen, Tuomas Hapola, Marian Heil, Andreas Maier, Jennifer Smillie
 *  \date      2019
 *  \copyright GPLv2 or later
 */

#pragma once

#include <memory>

#include <boost/core/demangle.hpp>

#include "fastjet/PseudoJet.hh"

namespace HEJ{

  //! Create a std::unique_ptr to a T object
  /**
   *  For non-array types this works like std::make_unique,
   *  which is not available under C++11
   */
  template<class T, class... Args>
  std::unique_ptr<T> make_unique(Args&&... a){
    return std::unique_ptr<T>{new T{std::forward<Args>(a)...}};
  }

  //! Create an array containing the passed arguments
  template<typename T, typename... U>
  constexpr
  std::array<T, 1 + sizeof...(U)> make_array(T t, U&&... rest){
    return {{t, std::forward<U>(rest)...}};
  }


  inline
  std::string join(
      std::string const & /* delim */
  ){
    return "";
  }

  inline
  std::string join(
      std::string const & /* delim */, std::string const & str
  ){
    return str;
  }

  //! Join strings with a delimiter
  /**
   *   @param delim      Delimiter to be put between consecutive strings
   *   @param first      First string
   *   @param second     Second string
   *   @param rest       Remaining strings
   */
  template<typename... Strings>
  std::string join(
      std::string const & delim,
      std::string const & first, std::string const & second,
      Strings&&... rest
  ){
    return join(delim, first + delim + second, std::forward<Strings>(rest)...);
  }

  //! Return the name of the argument's type
  template<typename T>
  std::string type_string(T&&){
    return boost::core::demangle(typeid(T).name());
  }

  //! Eliminate compiler warnings for unused variables
  template<typename... T>
  constexpr void ignore(T&&...) {}

  //! Check whether two doubles are closer than ep > 0 to each other
  inline
  bool nearby_ep(double a, double b, double ep){
    assert(ep > 0);
    return std::abs(a-b) < ep;
  }

  //! Check whether all components of two PseudoJets are closer than ep to each other
  inline
  bool nearby_ep(
      fastjet::PseudoJet const & pa, fastjet::PseudoJet const & pb,
      double ep
  ){
    assert(ep > 0);
    for(size_t i = 0; i < 4; ++i){
      if(!nearby_ep(pa[i], pb[i], ep)) return false;
    }
    return true;
  }

  inline
  bool nearby(
      fastjet::PseudoJet const & pa, fastjet::PseudoJet const & pb, double const norm = 1.
  ){
    return nearby_ep(pa, pb, 1e-7*norm);
  }

}
