/**
 *  \authors   Jeppe Andersen, Tuomas Hapola, Marian Heil, Andreas Maier, Jennifer Smillie
 *  \date      2019
 *  \copyright GPLv2 or later
 */
#include "HEJ/LesHouchesWriter.hh"

#include <cassert>
#include <utility>
#include <vector>

#include "HEJ/Event.hh"
#include "HEJ/event_types.hh"
#include "HEJ/utility.hh"

namespace HEJ{
  namespace{
    template<class T, class... Args>
      std::unique_ptr<T> make_unique(Args&&... a){
      return std::unique_ptr<T>{new T{std::forward<Args>(a)...}};
    }
  }

  LesHouchesWriter::LesHouchesWriter(
      std::string const & file, LHEF::HEPRUP heprup
  ):
    out_{file, std::fstream::in | std::fstream::out | std::fstream::trunc},
    writer_{HEJ::make_unique<LHEF::Writer>(out_)}
  {
    if(! out_.is_open()){
      throw std::ios_base::failure("Failed to open " + file);
    };
    writer_->heprup = std::move(heprup);
    // lhe Stardard: IDWTUP (negative => weights = +/-)
    // 1: weighted events, xs = mean(weight), XMAXUP given
    // 2: weighted events, xs = XSECUP, XMAXUP given
    // 3: unweighted events, no additional information given
    // 4: unweighted events, xs = mean(weight), no additional information given
    writer_->heprup.IDWTUP = 2;

    // use placeholders for unknown init block values
    // we can overwrite them after processing all events
    writer_->heprup.XSECUP = std::vector<double>(event_type::last_type+1, 0.);
    writer_->heprup.XERRUP = std::vector<double>(event_type::last_type+1, 0.);
    writer_->heprup.XMAXUP = std::vector<double>(event_type::last_type+1, 0.);
    write_init();
  }

  void LesHouchesWriter::write(Event const &  ev){
    assert(writer_ && out_.is_open());

    const double wt = ev.central().weight;
    writer_->hepeup = HEJ::to_HEPEUP(std::move(ev), &heprup());
    writer_->writeEvent();
    heprup().XSECUP[ev.type()] += wt;
    heprup().XERRUP[ev.type()] += wt*wt;
    if(wt > heprup().XMAXUP[ev.type()]){
      heprup().XMAXUP[ev.type()] = wt;
    }
  }

  // this function is called after overwritting the Les Houches init block
  // assert that we have overwritten *exactly* the init block,
  // i.e. an intact event block is next
  void assert_next_event_intact(std::istream & out){
    (void) out; // suppress compiler warnings if not in debug mode
#ifndef NDEBUG
    std::string line;
    getline(out, line);
    assert(line == "<event>");
#endif
  }

  void LesHouchesWriter::rewrite_init(){
    assert(writer_ && out_.is_open());

    // replace placeholder entries
    const auto pos = out_.tellp();
    out_.seekp(0);
    writer_->init();
    assert_next_event_intact(out_);
    out_.seekp(pos);
  }

  LesHouchesWriter::~LesHouchesWriter(){
    assert(writer_ && out_.is_open());

    for(auto & xs_err: heprup().XERRUP)
    {
      xs_err = sqrt(xs_err);
    }
    rewrite_init();
  }

}
